"""Asynchronous Python client for Klyqa Lights."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import Optional

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializationStrategy


class IntegerIsBoolean(SerializationStrategy):
    """Boolean serialization strategy for integers."""

    def serialize(self, value: bool) -> int:  # noqa: FBT001
        """Serialize a boolean to an integer."""
        return int(value)

    def deserialize(self, value: int) -> bool:
        """Deserialize an integer to a boolean."""
        return bool(value)


class BaseModel(DataClassORJSONMixin):
    """Base model for all Klyqa models."""

    # pylint: disable-next=too-few-public-methods
    class Config(BaseConfig):
        """Mashumaro configuration."""

        omit_none = True
        serialization_strategy = {bool: IntegerIsBoolean()}  # noqa: RUF012
        serialize_by_alias = True


@dataclass
class Wifi(BaseModel):
    """Object holding the Klyqa device Wi-Fi information.

    This object holds wireles information about the Klyqa device.

    Attributes
    ----------
        security: The security setting of the Wi-Fi network connected.
        rssi: The signal strength in dBm of the Wi-Fi network connected.
        channel: The channel of the Wi-Fi network the device is connected to.

    """

    security: str
    rssi: int
    channel: int


@dataclass
class ChipInfo(BaseModel):
    """Object holding the Klyqa device chip information.

    This object holds information about the Klyqa device chip.

    Attributes
    ----------
        version: The version of the chip
        model: The model of the chip.
        features: The chip features.

    """

    version: str
    model: str
    features: str


@dataclass
class CloudInfo(BaseModel):
    """Object holding the Klyqa device cloud information.

    This object holds information about the Klyqa device cloud
    connection.

    Attributes
    ----------
        host: The host the device connects to
        port: The port the device connects to
        enabled: The flag to disable the cloud feature.

    """

    host: str
    port: int
    enabled: bool


@dataclass
class GeoInfo(BaseModel):
    """Object holding the Klyqa device geo information.

    This object holds information about the Klyqa device geo
    settings.

    Attributes
    ----------
        timezone: The Posix timezone string
        lat: The latitude of the device
        lon: The longitude of the device.

    """

    timezone: str
    lat: float
    lon: float

@dataclass
class LocalInfo(BaseModel):
    """Object holding the Klyqa Light device information.

    This object holds information about the Klyqa Light.

    Attributes
    ----------
        ip_address: IP Address of the device.
        rest_port: Port of the REST interface
        ctrl_port: Ctrl port of the REST interface (not used)
    """
    ip_address: str
    rest_port: int
    ctrl_port: int

@dataclass
# pylint: disable-next=too-many-instance-attributes
class Info(BaseModel):
    """Object holding the Klyqa Light device information.

    This object holds information about the Klyqa Light.

    Attributes
    ----------
        firmware_date: Date where this firmware was build.
        firmware_branch: A string with the branch of the firmware.
        firmware_version: String containing the firmware version.
        sdk_version: String containing the SDK version.
        hardware_revision: An integer indicating the board revision.
        product_id: The product name.
        product_name: The friendly product name.
        device_id: Device ID of the Klyqa Light.
        service_name: Service name of the Klyqa Light.
        device_group: Device group string of the Klyqa Light.
        device_order: Device order number of the Klyqa Light.
        chip_info: Chip information of the Klyqa Light.
        cloud: Cloud information of the Klyqa Light.
        geo: Geo information of the Klyqa Light.

    """

    firmware_version: str = field(metadata=field_options(alias="app_ver"))
    firmware_date: str = field(metadata=field_options(alias="build_date"))
    firmware_branch: str = field(metadata=field_options(alias="branch"))
    sdk_version: str = field(metadata=field_options(alias="sdk_ver"))
    hardware_revision: int = field(metadata=field_options(alias="hw_revision"))
    product_id: str = field(metadata=field_options(alias="product_id"))
    device_id: str = field(metadata=field_options(alias="device_id"))
    service_name: str = field(metadata=field_options(alias="service_name"))
    chip_info: ChipInfo = field(metadata=field_options(alias="chip_info"))

    product_name: str | None = field(metadata=field_options(alias="product_name"), default="Klyqa Device")

class PowerOnBehavior(IntEnum):
    """Enum for the power on behavior of the Klyqa Light."""

    UNKNOWN = 0
    RESTORE_LAST = 1
    USE_DEFAULTS = 2


@dataclass
class FadeDuration:
    """Class to store switch fading duration values."""

    fadein: int = field(metadata=field_options(alias="in"))
    fadeout: int = field(metadata=field_options(alias="out"))


@dataclass
class RGBColor:
    """Class to store RGB color values."""

    red: int  # Value between 0 and 255
    green: int  # Value between 0 and 255
    blue: int  # Value between 0 and 255

    def __post_init__(self):
        # Ensure RGB values are within range.
        assert 0 <= self.red <= 255, "Red value must be between 0 and 255"
        assert 0 <= self.green <= 255, "Green value must be between 0 and 255"
        assert 0 <= self.blue <= 255, "Blue value must be between 0 and 255"


@dataclass
class Brightness:
    """Class to store Brgihtness percentage value."""

    percentage: int  # Value between 0 and 100

    def __post_init__(self) -> None:
        assert 0 <= self.percentage <= 100, "Red value must be between 0 and 255"


@dataclass
class ExternalConfig:
    """Class to store External Configuration values."""

    port: int
    channel: int
    universe: int
    autoreset: int
    mode: str


@dataclass
class Settings(BaseModel):
    """Object holding the Klyqa Light settings

    Attributes
    ----------

    """
    device_order: int = field(metadata=field_options(alias="order"))
    device_name: str = field(metadata=field_options(alias="device_name"))
    
    cloud: CloudInfo = field(metadata=field_options(alias="cloud"))
    local: LocalInfo = field(metadata=field_options(alias="local"))
    geo: GeoInfo = field(metadata=field_options(alias="geo"))

    device_group: str | None = field(
        metadata=field_options(alias="group"), default=None
    )

@dataclass
class State(BaseModel):
    """Object holding the Klyqa Light state.

    Represents a visible state of an Klyqa Light.

    Attributes
    ----------
        on: A boolean indicating the if the light if on or off.
        brightness: An integer between 0 and 100, representing the brightness.
        rgb: A tuple representing RGB color values.
        temperature: An integer representing the color temperature in mireds.

    """

    on: str = field(metadata=field_options(alias="status"))
    mode: str
    brightness: Brightness
    power_on_behaviour: int = field(metadata=field_options(alias="power_on"))
    active_command: int
    active_scene: str

    wifi_parameters: Wifi
    color: Optional[RGBColor] = None
    fade: FadeDuration | None = field(default=None)
    external: ExternalConfig | None = field(default=None)
    temperature: int | None = field(default=None)
    open_slots: int | None = field(default=None)



import logging

import aiohttp

from .exceptions import KlyqaConnectionError, KlyqaDeviceNotFoundError, KlyqaError

_LOGGER = logging.getLogger(__name__)
# _LOGGER.setLevel(logging.DEBUG)

class KlyqaCloud:
    BASE_URL = "https://app-api.prod.qconnex.io"

    def __init__(self, email, password, base_url=None) -> None:
        self.email = email
        self.password = password
        self.api_key = None
        self.base_url = base_url if base_url else self.BASE_URL
        self.session = aiohttp.ClientSession()

    async def login(self) -> str:
        """Login to the cloud and retrieve the API key."""
        login_url = self.base_url + "/auth/login"
        _LOGGER.debug("Trying to login at %s", login_url)
        payload = {"email": self.email, "password": self.password}

        async with self.session.post(login_url, json=payload, timeout=10) as resp:
            if resp.status == 201:
                data = await resp.json()
                self.api_key = data.get("accountToken")
                _LOGGER.debug("Login successful, API Key: %s", self.api_key)
                return self.api_key
            raise KlyqaConnectionError(f"Login failed with status code: {resp.status}")

    def get_api_key(self) -> str:
        """Get the API Key."""
        return self.api_key

    async def get_devices(self) -> str:
        """Fetch devices from the cloud."""
        if not self.api_key:
            raise KlyqaError("Not logged in. Call login() first.")

        devices_url = self.base_url + "/settings"
        headers = {"Authorization": f"Bearer {self.api_key}"}

        async with self.session.get(devices_url, headers=headers) as resp:
            if resp.status == 200:
                data = await resp.json()
                devices = data.get("devices", [])
                return devices
            raise KlyqaError(f"Failed to fetch devices with status code: {resp.status}")

    async def get_device_access_token(self, local_device_id) -> str:
        """Retrieve access token for a specific local device."""
        devices = await self.get_devices()
        for device in devices:
            if device.get("localDeviceId") == local_device_id:
                return device.get("accessToken")
        raise KlyqaDeviceNotFoundError(f"Device with localDeviceId {local_device_id} not found.")

    async def get_device_name(self, local_device_id) -> str:
        """Retrieve access token for a specific local device."""
        devices = await self.get_devices()
        for device in devices:
            if device.get("localDeviceId") == local_device_id:
                return device.get("name")
        raise KlyqaDeviceNotFoundError(f"Device with localDeviceId {local_device_id} not found.")

    async def close(self) -> None:
        await self.session.close()



import asyncio
import logging
import socket
from typing import Any, TypedDict

from aiohttp.client import ClientError, ClientSession
from aiohttp.hdrs import METH_GET, METH_PUT
from yarl import URL
import json

from .exceptions import KlyqaConnectionError, KlyqaAuthenticationError, KlyqaError
from .models import Info, RGBColor, State, Settings  # , PowerOnBehavior

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.DEBUG)

# console_handler = logging.StreamHandler()
# console_handler.setLevel(logging.DEBUG)

# formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
# console_handler.setFormatter(formatter)

# if not _LOGGER.handlers:
#     _LOGGER.addHandler(console_handler)


class KlyqaDevice:
    BASE_URL_TEMPLATE = "http://{ip}:{port}/api/v1/"

    host: str
    port: int = 3333
    request_timeout: int = 8
    access_token: str
    session: ClientSession | None = None

    _close_session: bool = False

    def __init__(self, host, port, access_token, session) -> None:
        """Init for Klyqa devices"""
        self.host = host
        self.port = port
        self.access_token = access_token

        self.session = session
        self.base_url = self.BASE_URL_TEMPLATE.format(ip=self.host, port=self.port)
        self.headers = {"Authorization": f"{self.access_token}"}
        self._state = {}  # Shadow state to store the device state

        _LOGGER.info(f"Device access token {self.access_token}")

        # Update the shadow state on initialization
        asyncio.create_task(self._update_state())

    async def _send_request(self, method, endpoint, json_payload=None) -> str:
        """Helper function to send HTTP requests."""
        url = self.base_url + endpoint
        async with self.session.request(
            method, url, json=json_payload, headers=self.headers
        ) as resp:
            if resp.status == 200:
                return await resp.json()
            if resp.status == 401:
                raise KlyqaAuthenticationError(
                    f"Request failed with status code {resp.status} for endpoint {endpoint}"
                )
            raise KlyqaError("Error in sending request to device")

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        data: dict[str, Any] | None = None,
    ) -> str:
        """Handle a request to a Klyqa Light device.

        A generic method for sending/handling HTTP requests done against
        the Klyqa Light API.

        Args:
        ----
            uri: Request URI, without '/api/v1/', for example, 'info'
            method: HTTP Method to use.
            data: Dictionary of data to send to the Klyqa Light.

        Returns:
        -------
            A Python string (JSON) with the response from the Klyqa Light API.

        Raises:
        ------
            KlyqaConnectionError: An error occurred while communicating with
                the Klyqa Light.
            KlyqaError: Received an unexpected response from the Klyqa Light
                API.

        """
        url = URL.build(
            scheme="http",
            host=self.host,
            port=self.port,
            path="/api/v1/",
        ).join(URL(uri))

        headers = {
            "Authorization": self.access_token,
            "User-Agent": "PythonKlyqa",
            "Accept": "application/json, text/plain, */*",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    json=data,
                    headers=headers,
                )
                response.raise_for_status()
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to Klyqa Light device"
            raise KlyqaConnectionError(msg) from exception
        except (
            ClientError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with Klyqa Light device"
            raise KlyqaConnectionError(msg) from exception

        return await response.text()

    async def _request_command(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        data: dict[str, Any] | None = None,
    ) -> str:
        """Handle a command request to a Klyqa Light device.

        A generic method for sending/handling HTTP command requests done against
        the Klyqa Light API.

        Args:
        ----
            uri: Request URI, without '/api/v1/', for example, 'info'
            method: HTTP Method to use.
            data: Dictionary of data to send to the Klyqa Light.

        Returns:
        -------
            A Python string (JSON) with the response from the Klyqa Light API.

        Raises:
        ------
            KlyqaConnectionError: An error occurred while communicating with
                the Klyqa Light.
            KlyqaError: Received an unexpected response from the Klyqa Light
                API.

        """
        url = URL.build(
            scheme="http",
            host=self.host,
            port=self.port,
            path="/api/v1/",
        ).join(URL(uri))

        headers = {
            "Authorization": self.access_token,
            "User-Agent": "PythonKlyqa",
            "Accept": "application/json, text/plain, */*",
        }

        send_data = {"command": data}

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    json=send_data,
                    headers=headers,
                )
                response.raise_for_status()
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to Klyqa Light device"
            raise KlyqaConnectionError(msg) from exception
        except (
            ClientError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with Klyqa Light device"
            raise KlyqaConnectionError(msg) from exception

        return await response.text()

    async def _update_state(self) -> None:
        """Update the shadow state by fetching the current state from the device."""
        self._state = await self._send_request(
            "PUT", "system/command", json_payload={"command": {"type": "request"}}
        )
        _LOGGER.info("Device state updated!")
        _LOGGER.debug("Device state: %s", self._state)

    async def info(self) -> Info:
        """Get devices information from Klyqa Light device.

        Returns
        -------
            A Info object, with information about the Klyqa Light device.

        """
        data = await self._request("system/info")
        return Info.from_json(data)
    
    async def settings(self) -> Settings:
        """Get device settings from Klyqa Light device.

        Returns
        -------
            A Info object, with settings from the Klyqa Light device.

        """
        data = await self._request("system/settings")
        return Settings.from_json(data)

    async def state(self) -> State:
        """Get the current state of Klyqa Light device.

        Returns
        -------
            A State object, with the current Klyqa Light state.

        """
        data = await self._request_command(
            "system/command", method=METH_PUT, data={"type": "request"}
        )
        # pylint: disable-next=no-member
        return State.from_json(data)
    
    async def change_device_name(self, new_name) -> None:
        """Change the name of the Klyqa Light device."""
        await self._request(
            "system/settings", method=METH_PUT, data={"type": "device_name", "name": new_name}
        )

    # pylint: disable-next=too-many-arguments
    async def light(
        self,
        *,
        on: bool | None = None,
        brightness: int | None = None,
        color: RGBColor | None = None,
        temperature: int | None = None,
    ) -> None:
        """Change state of an Klyqa Light device.

        Args:
        ----
            on: A boolean, true to turn the light on, false otherwise.
            brightness: The brightness of the light, between 0 and 255.
            color: RGBColor
            temperature: The white temperature of the light, in kelvin.

        Raises:
        ------
            KlyqaError: The provided values are invalid.

        """
        if temperature and (color):
            msg = "Cannot set temperature together with color"
            raise KlyqaError(msg)

        class LightState(TypedDict, total=False):
            """Describe state dictionary that can be set on a light."""

            brightness: int
            color: RGBColor
            status: str
            temperature: int

        state: LightState = {}

        state["type"] = "request"

        if on is not None:
            if on:
                state["status"] = "on"
            else:
                state["status"] = "off"

        if brightness is not None:
            if not 0 <= brightness <= 100:
                msg = "Brightness not between 0 and 100"
                raise KlyqaError(msg)
            state["brightness"] = {"percentage": brightness}

        if color is not None:
            state["color"] = {
                "red": color.red,
                "green": color.green,
                "blue": color.blue,
            }

        if temperature is not None:
            if not 2200 <= temperature <= 7500:
                msg = "White temperature out of range"
                raise KlyqaError(msg)
            state["temperature"] = temperature

        if not state:
            msg = "No parameters to set, light not adjusted"
            raise KlyqaError(msg)

        _LOGGER.info(f"Device state {state}")

        await self._request_command(
            "system/command",
            method=METH_PUT,
            data=state,
        )

    async def close(self) -> None:
        """Close the session."""
        await self.session.close()



"""Exceptions for Klyqa Lights."""


class KlyqaError(Exception):
    """Generic Klyqa Light exception."""


class KlyqaConnectionError(KlyqaError):
    """Klyqa Light connection exception."""


class KlyqaAuthenticationError(KlyqaError):
    """Klyqa Light Authentication exception."""

class KlyqaDeviceNotFoundError(KlyqaError):
    """Klyqa Cloud device not found exception."""



# __init__.py
from .cloud import KlyqaCloud
from .device import KlyqaDevice
from .exceptions import KlyqaError, KlyqaConnectionError, KlyqaAuthenticationError
from .models import Info, State, Settings

__all__ = ['KlyqaCloud', 'KlyqaDevice', 'KlyqaError', 'KlyqaConnectionError', 'KlyqaAuthenticationError', 'Info', 'Settings', 'State']